/*
 * mat.c
 *
 * Let us know about IM speed
 * Compare with libckpt, application MAT
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#include <util.h>
void checkpoint_here(void);

ulong freq;

/*
 * timeval_subtract
 * (cogido del info de libc)
 */
     /* Subtract the `struct timeval' values X and Y,
        storing the result in RESULT.
        Return 1 if the difference is negative, otherwise 0.  */
     
     int
     timeval_subtract (struct timeval *result, 
                       struct timeval *x, 
                       struct timeval *y)
     {
       /* Perform the carry for the later subtraction by updating Y. */
       if (x->tv_usec < y->tv_usec) {
         int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
         y->tv_usec -= 1000000 * nsec;
         y->tv_sec += nsec;
       }
       if (x->tv_usec - y->tv_usec > 1000000) {
         int nsec = (x->tv_usec - y->tv_usec) / 1000000;
         y->tv_usec += 1000000 * nsec;
         y->tv_sec -= nsec;
       }
     
       /* Compute the time remaining to wait.
          `tv_usec' is certainly positive. */
       result->tv_sec = x->tv_sec - y->tv_sec;
       result->tv_usec = x->tv_usec - y->tv_usec;
     
       /* Return 1 if result is negative. */
       return x->tv_sec < y->tv_sec;
     }

/*
 * matread()
 */
int
matread(char *path, void *addr, int N)
{
	int fd, r;
	uint sz = N*N*sizeof(int);

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		perror("open");
		exit(1);
	}
	r = read(fd, addr, sz); 
	if (r == -1) {
		perror("read");
		exit(1);
	} else if (r != sz) {
		fprintf(stderr,
                        "Warning: %u of %u bytes read!\n",
		        r, sz);
	}
	close(fd);
	return(0);
}

/*
 * matwrite()
 */
int
matwrite(char *path, void *addr, int N)
{
	int fd, r;
	uint sz = N*N*sizeof(int);

	fd = open(path, O_WRONLY|O_CREAT, S_IREAD|S_IWRITE);
	if (fd == -1) {
		perror("open");
		exit(1);
	}
	r = write(fd, addr, sz); 
	if (r == -1) {
		perror("write");
	exit(1);
	} else if (r != sz) {
		fprintf(stderr,
                        "Warning: %u of %u bytes written!\n",
		        r, sz);
	}
	close(fd);
	return(0);
}

/*
 * matmul()
 *	b = a * a
 */
void
matmul(int *A, int *A2, int *B, int N)
{
	uint i, j, k;

	for (i=0; i<N; ++i) {
		for (j=0; j<N; ++j) {
			for (k=0; k<N; ++k) {
				*B += *(A+i*N+k) * *(A2+k*N+j);

			}
			B++;
		}
		if ( (i) % freq == 0 ) {
			checkpoint_here();
			printf("Multiplicada fila %ud\n",i);
		}
	}
}

/*
 * main()
 */
int
ckpt_target(int argc, char **argv)
{
	ulong matsz,npgs; 
	void *mem;
	void *A1, *A2, *B;
	struct timeval t, t1, t11, t22;

	/*
	 * Process arguments
	 */	
	if (argc < 3) {
		printf("Usage: %s N F\n", argv[0]);
		printf("\tN : # of rows and columns for square matrices\n");
		printf("\tF : Checkpoint frequency\n");
		exit(1);
	}

	matsz = atoi(argv[1]);
	if (matsz == 0) {
		printf("N must be > 0\n");
		exit(1);
	}
	freq = atoi(argv[2]);
	printf("Matrix size: %ldx%ld -- Chkpt every %ld rows\n",matsz,matsz,freq);
 

	/* region must hold 3 matsz x matsz matrices*/
	npgs = btop(3 * sizeof(int) * matsz * matsz + (NBPG - 1));

	printf("matsz %lu, npgs %lu (0x%lx)\n", matsz, npgs, npgs);

	/*
	 * Get a new virtual memory region of appropiate size
	 */
	printf("Allocating memory...\n");
	mem = malloc(ptob(npgs));
	if (mem == NULL) {
		perror("malloc");
		return(1);
	}
	printf("...memory successfully allocated at %p\n", mem);

	/*
	 * Run MAT 
	 */
	A1 = (void *)mem;
	A2 = (void *)(mem + matsz*matsz*sizeof(int));
	B = (void *)(mem + 2*matsz*matsz*sizeof(int));

	printf("Initializing MATs...\n");
	gettimeofday(&t1, NULL);
	if (matread("A1.mat", A1, matsz) != 0) {
		fprintf(stderr, "Can't read A1 matrix\n");
		return 1;
	}
	if (matread("A2.mat", A2, matsz) != 0) {
		fprintf(stderr, "Can't read A2 matrix\n");
		return 1;
	}
	gettimeofday(&t11, NULL);
	timeval_subtract(&t, &t11, &t1);
	printf("time to read: %lu s, %lu us\n", t.tv_sec, t.tv_usec);

	printf("Multiplying MATs...\n");
	gettimeofday(&t11, NULL);
	matmul(A1, A2, B, matsz);
	gettimeofday(&t22, NULL);
	timeval_subtract(&t, &t22, &t11);
	printf("time to mul: %lu s, %lu us\n", t.tv_sec, t.tv_usec);

	printf("Writing results...\n");
	gettimeofday(&t11, NULL);
	matwrite("B.mat", B, matsz);
	gettimeofday(&t22, NULL);
	timeval_subtract(&t, &t22, &t11);
	printf("time to write: %lu s, %lu us\n", t.tv_sec, t.tv_usec);
	timeval_subtract(&t, &t22, &t1);
	printf("time elapsed: %lu s, %lu us\n", t.tv_sec, t.tv_usec);

	printf("Bye!\n");
	
	return(0);
}


