/*
 * matprint.c
 *	Print matrices. 
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/mman.h>


#include <util.h>

/*
 * matprint()
 */
void
matprint(char *title, int *A, int N)
{
	uint i, j;

	printf("matrix: '%s'\n", title);
	for (j=0; j<N; j++) {
		for (i=0; i<N; i++) {
			printf("%d ", *A++);
		}
		printf("\n");
	}
}

/*
 * matread()
 */
int
matread(char *path, void *addr, int N)
{
	int fd, r;
	uint sz = N*N*sizeof(int);

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		perror("open");
		exit(1);
	}
	r = read(fd, addr, sz); 
	if (r == -1) {
		perror("read");
		exit(1);
	} else if (r != sz) {
		fprintf(stderr,
		        "Warning: read %u of %u bytes!\n",
			r, sz);
	}
	close(fd);
	return(0);
}

/*
 * main()
 */
int
main(int argc, char **argv)
{
	ulong matsz, npgs; 
	char *matname;
	void *mem;

	/*
	 * Process arguments
	 */	
	if (argc < 3) {
		printf("Usage: %s N file_name\n", argv[0]);
		printf("\tN: number of rows and columns for square matrix\n");
		exit(1);
	}
	argv++; argc--; 
	matsz = atoi(*argv);
	if (matsz == 0) {
		printf("N must be > 0\n");
		exit(1);
	}
	argv++; argc--; 
	matname = *argv;

	/* region must hold matsz x matsz matrices*/
	npgs = btop(sizeof(int) * matsz * matsz + (NBPG - 1));

	printf("matsz %lu, npgs %lu (0x%lx)\n", matsz, npgs, npgs);

	/*
	 * Get a new virtual memory region of appropiate size
	 */
	printf("Allocating memory...\n");
	mem = (void *)mmap(0, ptob(npgs), PROT_READ|PROT_WRITE,
		MAP_ANON|MAP_PRIVATE, 0, 0);
	if (!mem) {
		perror("mmap");
		return(1);
	}
	printf("...memory successfully allocated at %p\n", mem);

	/*
	 * Read MAT 
	 */
	printf("Reading MAT...\n");
	if (matread(matname, mem, matsz) != 0) {
		fprintf(stderr, "Error reading matrix\n");
		return 1;
	}
	printf("Printing MAT...\n");
	matprint(matname, mem, matsz);

	printf("Bye!\n");
	
	return(0);
}


